/* ------------------------------------------------------------------------ */
/*                                                                          */
/*  Copyright (C)                                                           */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful, but     */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-    */
/*  ABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General        */
/*  Public License for more details.                                        */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                   */
/*                                                                          */
/* ------------------------------------------------------------------------ */
/*  $Author: vogje02 $ $Revision: 1.2 $ $Date: 2002/05/18 11:27:13 $      */
/* ------------------------------------------------------------------------ */

#include <LogWriter.h>

/**
 * Constructor
 */
CLogWriter::CLogWriter( char* file_name, int dl )
{
  // The the file name and open the file
  if( file_name != NULL ) {
    strcpy( fileName, file_name );
    log_file = fopen( file_name, "w+" );
  } else
    log_file = stderr;
  
  // Set defaults
  debug_level = dl;
  
  // Reset the size, maxSize is 10MB, count is 5 -> max. 5 log file, each 5MB.
  size = 5;
  currentLogCount = 0;
  logCount = 5;
  maxSize = size * 1024 * 1024;
  
  log( INFO, "CLogWriter", "Constructor", "Logger initialized, level: %d count: %d size: %d MB", 
       debug_level, logCount, size );
}

/**
 * Constructor
 */
CLogWriter::CLogWriter( char* file_name, int dl, int count, int inSize )
{
  if( file_name != NULL ) {
    strcpy( fileName, file_name );
    log_file = fopen( file_name, "w+" );
  } else
    log_file = stderr;
  
  debug_level = dl;

  // Reset the size, maxSize is 10MB, count is 5 -> max. 5 log file, each 5MB.
  size = inSize;
  currentLogCount = 0;
  logCount = count;
  maxSize = inSize * 1024 * 1024;
  
  log( INFO, "CLogWriter", "Constructor", "Initialized - level: %d count: %d size: %d MB", debug_level, logCount, size );
}

/**
 * Destructor
 */
CLogWriter::~CLogWriter()
{
  fflush( log_file );
  fclose( log_file );
}

/**
 * Logging routine.
 */
void CLogWriter::log( int level, char* cl, char* method, char* fmt, ... )
{
  // If severity is above the debug level exit immediately 
  if( level > debug_level )
    return;
  
  char msg[1024];
  char buffer[1024];
  char stime[40];
  va_list marker;

  *buffer = '\0';
  *msg = '\0';
  // Initialize variable arguments. 
  va_start( marker, fmt );     
  vsprintf( msg, fmt, marker ); 
  
  // Get the current time
  GetTime( stime );

  // Switch according to severity 
  switch( level ) {
    
  case FATAL:
    sprintf( buffer, "%s - [F] - %s::%s - %s", stime, cl, method, msg );
    break;
    
  case ERROR:
    sprintf( buffer, "%s - [E] - %s::%s - %s", stime, cl, method, msg );
    break;
    
  case INFO:
    sprintf( buffer, "%s - [I] - %s::%s - %s", stime, cl, method, msg );
    break;

  case WARNING:
    sprintf( buffer, "%s - [W] - %s::%s - %s", stime, cl, method, msg );
    break;

  case DEBUG0:
    sprintf( buffer, "%s - [0] - %s::%s - %s", stime, cl, method, msg );
    break;

  case DEBUG1:
    sprintf( buffer, "%s - [1] - %s::%s - %s", stime, cl, method, msg );
    break;

  case DEBUG2:
    sprintf( buffer, "%s - [2] - %s::%s - %s", stime, cl, method, msg );
    break;

  default:
    sprintf( buffer, "%s - [-] - %s::%s - %s", stime, cl, method, msg );
    break;
  }
  
  // Increment the size
  size += ( strlen( buffer ) + 1 );
  
  // Print the og message
  fprintf( log_file, "%s\n", buffer );
  fflush( log_file );
  
  // Rotate the log file if size is reached
  if( size > maxSize ) {
    fprintf( log_file, "MAX log size reached. Will rotate now." );
    Rotate();
    size = 0;
  }
 
  // Cleanup
  va_end( marker );
}

/**
 * Get the current system time
 */
void CLogWriter::GetTime( char* stime )
{
  SYSTEMTIME ti;
  GetLocalTime( &ti );
  sprintf( stime, "%02d/%02d/%4d %02d:%02d:%02d.%03d", ti.wDay, ti.wMonth, ti.wYear, ti.wHour, ti.wMinute, 
	   ti.wSecond, ti.wMilliseconds );
}

/**
 * Rotate the log files.
 */
void CLogWriter::Rotate()
{
  if( log_file == stderr )
    return;

  // Close the log file
  fclose( log_file );
  
  // Copy the log file to a file with name logFile.currentLogCount
  char newFileName[256];
  sprintf( newFileName, "%s.%d", fileName, currentLogCount );
  remove( newFileName );
  rename( fileName, newFileName );
  
  // Open a new file.
  log_file = fopen( fileName, "w+" );
  
  // Increase the current log counter and check against maximum number of log files.
  currentLogCount++;
  if( currentLogCount > logCount - 1 )
    currentLogCount = 0;
}
