/* ------------------------------------------------------------------------ */
/*                                                                          */
/*  Copyright (C)                                                           */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful, but     */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-    */
/*  ABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General        */
/*  Public License for more details.                                        */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                   */
/*                                                                          */
/* ------------------------------------------------------------------------ */
/*  $Author: vogje02 $ $Revision: 1.2 $ $Date: 2002/05/18 11:27:13 $      */
/* ------------------------------------------------------------------------ */

/**
 * This module contains functions that access objects in performance data.
 */
#include "perfdata.h"

// Extern function
extern DWORD  PX_PROCESS;
extern DWORD  PX_PROCESS_ID;
extern DWORD  PX_THREAD;
extern TCHAR  INDEX_PROCTHRD_OBJ[2*INDEX_STR_LEN];
extern void   SetPerfIndexes();

DWORD	GetTitleIdx (LPTSTR Title[], DWORD LastIndex, LPTSTR Name);

/**
 * Get the number of threads
 */
int ProcessGetNumThreads( PPERF_DATA data, unsigned int ParentIndex )
{		
  register				i = 0, k = 0;
  PPERF_OBJECT		tObject;
  PPERF_INSTANCE	tInstance;
  
  /*
   * Set first instance
   */
  tObject = FindObject (data, PX_THREAD);
  tInstance	= FirstInstance(tObject);
  
  /*
   * Find the right instance
   */
  while (tInstance && i < tObject->NumInstances) 
    {
      /*
       * Count number of threads
       */
      if (ParentIndex == tInstance->ParentObjectInstance)
	      k++;
      tInstance = NextInstance (tInstance);
      i++;	
    }
  
  return(k);
}

/**
 * Get the process ID from the performance data.
 */
int ProcessGetPid( PPERF_OBJECT pObject, PPERF_INSTANCE pInstance )
{
  DWORD           *pdwProcID;
  PPERF_COUNTER   pCounterProcID;
  
  // Find counter
  if (pCounterProcID = FindCounter (pObject, PX_PROCESS_ID)) {
    pdwProcID = (DWORD *) CounterData (pInstance, pCounterProcID);
    if(pdwProcID) return *pdwProcID; 
  }
  
  return (-1);
}

/**
 * Setup the perf data indexes.
 */
void SetPerfIndexes ()
{
	LPTSTR  TitleBuffer;
	LPTSTR  *Title;
	DWORD   Last;
	DWORD   dwR;
	HKEY		hPerfKey = HKEY_PERFORMANCE_DATA;
	HKEY		hMachineKey = HKEY_LOCAL_MACHINE;  // get title index from this key

  dwR = GetPerfTitleSz (hMachineKey, hPerfKey, &TitleBuffer, &Title, &Last);

  if (dwR != ERROR_SUCCESS) {
    fprintf( stderr, "Could not get performance indexes. Return Code %d", dwR );        
    return;
  }

  PX_PROCESS                       = GetTitleIdx (Title, Last, PN_PROCESS);
  PX_PROCESS_CPU                   = GetTitleIdx (Title, Last, PN_PROCESS_CPU);
  PX_PROCESS_PRIV                  = GetTitleIdx (Title, Last, PN_PROCESS_PRIV);
  PX_PROCESS_USER                  = GetTitleIdx (Title, Last, PN_PROCESS_USER);
  PX_PROCESS_WORKING_SET           = GetTitleIdx (Title, Last, PN_PROCESS_WORKING_SET);
  PX_PROCESS_PEAK_WS               = GetTitleIdx (Title, Last, PN_PROCESS_PEAK_WS);
  PX_PROCESS_PRIO                  = GetTitleIdx (Title, Last, PN_PROCESS_PRIO);
  PX_PROCESS_ELAPSE                = GetTitleIdx (Title, Last, PN_PROCESS_ELAPSE);
  PX_PROCESS_ID                    = GetTitleIdx (Title, Last, PN_PROCESS_ID);
  PX_PROCESS_PRIVATE_PAGE          = GetTitleIdx (Title, Last, PN_PROCESS_PRIVATE_PAGE);
  PX_PROCESS_VIRTUAL_SIZE          = GetTitleIdx (Title, Last, PN_PROCESS_VIRTUAL_SIZE);
  PX_PROCESS_PEAK_VS               = GetTitleIdx (Title, Last, PN_PROCESS_PEAK_VS);
  PX_PROCESS_FAULT_COUNT           = GetTitleIdx (Title, Last, PN_PROCESS_FAULT_COUNT);

	PX_THREAD                        = GetTitleIdx (Title, Last, PN_THREAD);
  PX_THREAD_CPU                    = GetTitleIdx (Title, Last, PN_THREAD_CPU);
  PX_THREAD_PRIV                   = GetTitleIdx (Title, Last, PN_THREAD_PRIV);
  PX_THREAD_USER                   = GetTitleIdx (Title, Last, PN_THREAD_USER);
  PX_THREAD_START                  = GetTitleIdx (Title, Last, PN_THREAD_START);
  PX_THREAD_SWITCHES               = GetTitleIdx (Title, Last, PN_THREAD_SWITCHES);
  PX_THREAD_PRIO                   = GetTitleIdx (Title, Last, PN_THREAD_PRIO);
  PX_THREAD_BASE_PRIO              = GetTitleIdx (Title, Last, PN_THREAD_BASE_PRIO);
  PX_THREAD_ELAPSE                 = GetTitleIdx (Title, Last, PN_THREAD_ELAPSE);

  wsprintf (INDEX_PROCTHRD_OBJ, TEXT("%ld %ld"), PX_PROCESS, PX_THREAD);

  LocalFree (TitleBuffer);
  LocalFree (Title);

}	


/**
 * Returns pointer to the first object in pData. If pData is NULL then NULL 
 * is returned.
 */
PPERF_OBJECT FirstObject (PPERF_DATA pData)
{
    if (pData)
        return ((PPERF_OBJECT) ((PBYTE) pData + pData->HeaderLength));
    else
        return NULL;
}

/**
 * Returns pointer to the next object following pObject.
 *
 * If pObject is the last object, bogus data maybe returned. The caller should 
 * do the checking. If pObject is NULL, then NULL is returned.
 */
PPERF_OBJECT NextObject (PPERF_OBJECT pObject)
{
  if (pObject)
    return ((PPERF_OBJECT) ((PBYTE) pObject + pObject->TotalByteLength));
  else
    return NULL;
}

/**
 * Returns pointer to object with TitleIndex.  If not found, NULL is 
 * returned.
 */
PPERF_OBJECT FindObject (PPERF_DATA pData, DWORD TitleIndex)
{
	PPERF_OBJECT pObject;
	DWORD        i = 0;

  if (pObject = FirstObject (pData))
    while (i < pData->NumObjectTypes) {
      if (pObject->ObjectNameTitleIndex == TitleIndex)
        return pObject;

      pObject = NextObject (pObject);
      i++;
  }

  return NULL;
}

/**
 * Get the Title indexes
 */
DWORD GetTitleIdx (LPTSTR Title[], DWORD LastIndex, LPTSTR Name)
{
	DWORD   Index;

  for (Index = 0; Index <= LastIndex; Index++)
    if (Title[Index])
      if (!lstrcmpi (Title[Index], Name))
        return Index;

  return 0;
}

LPTSTR      *gPerfTitleSz;
LPTSTR      TitleData;

/**
 * Get a new set of performance data.
 *
 * *ppData should be NULL initially. This function will allocate a buffer big 
 * enough to hold the data requested by szObjectIndex.
 *
 * *pDataSize specifies the initial buffer size.  If the size is too small, the 
 * function will increase it until it is big enough then return the size through 
 * *pDataSize. Caller should deallocate *ppData if it is no longer being used.
 */
DWORD GetPerfData(HKEY        hPerfKey,
		  		        LPTSTR      szObjectIndex,
                  PPERF_DATA  *ppData,
                  DWORD       *pDataSize)
{
	DWORD   DataSize;
	DWORD   dwR;
	DWORD   Type;

	if (!*ppData)
	{
    *ppData = (PPERF_DATA) LocalAlloc (LMEM_FIXED, *pDataSize);
	}
	
	do  
	{
		DataSize = *pDataSize;

		/*
		 * Try to get the performance data
		 */
		__try{
		 			dwR = RegQueryValueEx (hPerfKey, szObjectIndex, NULL, &Type, (BYTE *)*ppData, &DataSize);
		} __except( GetExceptionCode() == EXCEPTION_ACCESS_VIOLATION ) 
		{
			/*
			 * In case of access violation return with AWF_ERROR
			 */
			return(-1); 
		}		

    if (dwR == ERROR_MORE_DATA)
    {
      LocalFree (*ppData);
      *pDataSize += 1024;
      *ppData = (PPERF_DATA) LocalAlloc (LMEM_FIXED, *pDataSize);
    }

    if (!*ppData)
    {
      LocalFree (*ppData);
      return ERROR_NOT_ENOUGH_MEMORY;
     }

  } while (dwR == ERROR_MORE_DATA);

  return dwR;
}

#ifdef UNICODE

#define atoi    atoiW

/**
 * Unicode version of atoi.
 */
INT atoiW (LPTSTR s)
{
	INT i = 0;

	while (iswdigit (*s)) {
		i = i*10 + (BYTE)*s - L'0';
    s++;
  }

  return i;
}

#endif

/**
 * Retrieves the performance data title strings.
 *
 * Caller should provide two pointers, one for buffering the title strings the 
 * other for indexing the title strings.  This function will allocate memory for 
 * the TitleBuffer and TitleSz.  To get the title string for a particular title 
 * index one would just index the TitleSz. *TitleLastIdx returns the highest 
 * index can be used.  If TitleSz[N] is NULL then there is no Title for index N.
 */

DWORD   GetPerfTitleSz (HKEY    hKeyMachine, HKEY    hKeyPerf, LPTSTR  *TitleBuffer,
                        LPTSTR  *TitleSz[], DWORD   *TitleLastIdx)
{
	HKEY	  hKey1, hKey2;
	DWORD   Type, DataSize, dwR, Len, Index;
	LPTSTR  szCounterValueName, szTitle;

  /*
	 * Initialize
	 */
  hKey1        = NULL;
  hKey2        = NULL;
  *TitleBuffer = NULL;
  *TitleSz     = NULL;

  /* 
	 * Open the perflib key to find out the last counter's index and system version.
	 */
  dwR = RegOpenKeyEx (hKeyMachine, TEXT("software\\microsoft\\windows nt\\currentversion\\perflib"),
                      0, KEY_READ, &hKey1);
  if (dwR != ERROR_SUCCESS)
    goto done;

  /*
	 * Get the last counter's index so we know how much memory to allocate for TitleSz
	 */
  DataSize = sizeof (DWORD);
  dwR = RegQueryValueEx (hKey1, TEXT("Last Counter"), 0, &Type, (LPBYTE)TitleLastIdx, &DataSize);
  if (dwR != ERROR_SUCCESS)
    goto done;


  /*
	 * NT 1.0a or later.  Get the counters in key HKEY_PERFORMANCE_KEY
	 * and from value "Counter 009"
	 */
  szCounterValueName = TEXT("Counter 009");
  hKey2 = hKeyPerf;

  /*
	 * Find out the size of the data.
	 */
  dwR = RegQueryValueEx (hKey2, szCounterValueName, 0, &Type, 0, &DataSize);
  if (dwR != ERROR_SUCCESS)
    goto done;

  /*
	 * Allocate memory  
	 */
  *TitleBuffer = (LPTSTR)LocalAlloc (LMEM_FIXED, DataSize);
  if (!*TitleBuffer) {
    dwR = ERROR_NOT_ENOUGH_MEMORY;
    goto done;
  }

  *TitleSz = (LPTSTR *)LocalAlloc (LPTR, (*TitleLastIdx+1) * sizeof (LPTSTR));
  if (!*TitleSz) {
     dwR = ERROR_NOT_ENOUGH_MEMORY;
     goto done;
  }

  /*
	 * Query the data
	 */
  dwR = RegQueryValueEx (hKey2, szCounterValueName, 0, &Type, (BYTE *)*TitleBuffer, &DataSize);
  if (dwR != ERROR_SUCCESS)
    goto done;

  /*
	 * Setup the TitleSz array of pointers to point to beginning of each title string.
	 * TitleBuffer is type REG_MULTI_SZ.
	 */
  szTitle = *TitleBuffer;

  while (Len = lstrlen (szTitle)) {
    Index = atoi (szTitle);

    szTitle = szTitle + Len +1;

    if (Index <= *TitleLastIdx)
      (*TitleSz)[Index] = szTitle;

    szTitle = szTitle + lstrlen (szTitle) +1;
  }

done:

  /*
	 * Done. Now cleanup!
	 */
  if (dwR != ERROR_SUCCESS) 
	{
    /*
		 * There was an error, free the allocated memory    
		 */
    if (*TitleBuffer) LocalFree (*TitleBuffer);
    if (*TitleSz)     LocalFree (*TitleSz);
  }

  /*
	 * Close the hKeys.
	 */
  if (hKey1) RegCloseKey (hKey1);
  if (hKey2 && hKey2 != hKeyPerf) RegCloseKey (hKey2);

  return(dwR);
}


/**
 * Returns pointer to the first instance of pObject type. If pObject is NULL 
 * then NULL is returned.
 */
PPERF_INSTANCE   FirstInstance (PPERF_OBJECT pObject)
{
  if (pObject)
    return (PPERF_INSTANCE)((PCHAR) pObject + pObject->DefinitionLength);
  else
    return NULL;
}

/**
 * Returns pointer to the next instance following pInst.
 *
 * If pInst is the last instance, bogus data maybe returned. The caller should 
 * do the checking.
 */
PPERF_INSTANCE   NextInstance (PPERF_INSTANCE pInst)
{
	PERF_COUNTER_BLOCK *pCounterBlock;

  if (pInst) {
    pCounterBlock = (PERF_COUNTER_BLOCK *)((PCHAR) pInst + pInst->ByteLength);
    return (PPERF_INSTANCE)((PCHAR) pCounterBlock + pCounterBlock->ByteLength);
  } else
    return NULL;
}

/**
 * Returns the name of the pInst. If pInst is NULL then NULL is returned.
 */
LPTSTR  InstanceName (PPERF_INSTANCE pInst)
{
  if (pInst)
    return (LPTSTR)((PCHAR) pInst + pInst->NameOffset);
  else
    return NULL;
}

/**
 * Find the first counter in pObject. Returns a pointer to the first counter.
 * If pObject is NULL then NULL is returned.
 */
PPERF_COUNTER FirstCounter (PPERF_OBJECT pObject)
{
  if (pObject)
    return (PPERF_COUNTER)((PCHAR) pObject + pObject->HeaderLength);
  else
    return NULL;
}
	
/**
 * If pCounter is the last counter of an object type, bogus data maybe returned.
 * The caller should do the checking. Returns a pointer to a counter. If 
 * pCounter is NULL then NULL is returned.
 */
PPERF_COUNTER NextCounter (PPERF_COUNTER pCounter)
{
  if (pCounter)
    return (PPERF_COUNTER)((PCHAR) pCounter + pCounter->ByteLength);
  else
    return NULL;
}

/**
 * Find a counter specified by TitleIndex. Returns a pointer to the counter. If 
 * counter is not found then NULL is returned.
 */
PPERF_COUNTER FindCounter (PPERF_OBJECT pObject, DWORD TitleIndex)
{
PPERF_COUNTER pCounter;
DWORD         i = 0;

    if (pCounter = FirstCounter (pObject))
        while (i < pObject->NumCounters)
            {
            if (pCounter->CounterNameTitleIndex == TitleIndex)
                return pCounter;

            pCounter = NextCounter (pCounter);
            i++;
            }

    return NULL;

}

/**
 * Returns counter data for an object instance. If pInst or pCount is NULL then 
 * NULL is returne.
 */
PVOID CounterData (PPERF_INSTANCE pInst, PPERF_COUNTER pCount)
{
	PPERF_COUNTER_BLOCK pCounterBlock;

	if (pCount && pInst) {
		pCounterBlock = (PPERF_COUNTER_BLOCK)((PCHAR)pInst + pInst->ByteLength);
		return (PVOID)((PCHAR)pCounterBlock + pCount->CounterOffset);
	}
  else
    return NULL;
}
